﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Data;

namespace HeuristicLab.Problems.QuadraticAssignment {
  public static class MultidimensionalScaling {

    public static DoubleMatrix Classic(DoubleMatrix distances, out double error) {
      if (distances == null) throw new ArgumentNullException("distances");
      if (distances.Rows != distances.Columns) throw new ArgumentException("Distance matrix must be a square matrix.", "distances");
      error = 0.0;

      int dimension = distances.Rows;
      if (dimension == 1) return new DoubleMatrix(new double[,] { { 0, 0 } });
      else if (dimension == 2) return new DoubleMatrix(new double[,] { { 0, distances[0, 1] } });

      double[,] A = new double[dimension + 1, dimension + 1];
      for (int i = 1; i < dimension; i++) {
        for (int j = i + 1; j < dimension + 1; j++) {
          A[i, j] = -0.5 * distances[i - 1, j - 1] * distances[i - 1, j - 1];
          A[j, i] = A[i, j];
          A[0, 0] += A[i, j] / (dimension * dimension);
          double avg = A[i, j] / (double)dimension;
          A[0, j] += avg;
          A[0, i] += avg;
          A[i, 0] += avg;
          A[j, 0] += avg;
        }
      }
      double[,] B = new double[dimension, dimension];
      for (int i = 0; i < dimension - 1; i++) {
        for (int j = i + 1; j < dimension; j++) {
          B[i, j] = A[i + 1, j + 1] - A[0, j + 1] - A[i + 1, 0] + A[0, 0];
        }
      }
      double[] eigenvalues;
      double[,] eigenvectors;
      alglib.smatrixevd(B, dimension, 1, true, out eigenvalues, out eigenvectors);

      DoubleMatrix coordinates = new DoubleMatrix(dimension, 2);
      for (int i = 0; i < dimension; i++) {
        for (int j = 0; j < dimension; j++) {
          coordinates[i, 0] += eigenvectors[eigenvectors.GetLength(0) - 1, j]
                             * distances[i, j];
          coordinates[i, 1] += eigenvectors[eigenvectors.GetLength(0) - 2, j]
                             * distances[i, j];
        }
      }

      return coordinates;
    }

    /// <summary>
    /// Performs the Kruskal-Shepard algorithm and applies a gradient descent method
    /// to fit the coordinates such that the difference between the fit distances
    /// and the actual distances is minimal.
    /// </summary>
    /// <param name="distances">A symmetric NxN matrix that specifies the distances between each element i and j. Diagonal elements are ignored.</param>
    /// <param name="error">Returns the error between the fit distances and the actual distances.</param>
    /// <returns>A Nx2 matrix where the first column represents the x- and the second column the y coordinates.</returns>
    public static DoubleMatrix Metric(DoubleMatrix distances, out double error) {
      if (distances == null) throw new ArgumentNullException("distances");
      if (distances.Rows != distances.Columns) throw new ArgumentException("Distance matrix must be a square matrix.", "distances");
      error = 0.0;

      int dimension = distances.Rows;
      if (dimension == 1) return new DoubleMatrix(new double[,] { { 0, 0 } });
      else if (dimension == 2) return new DoubleMatrix(new double[,] { { 0, distances[0, 1] } });

      DoubleMatrix coordinates = new DoubleMatrix(dimension, 2);
      double rad = (2 * Math.PI) / coordinates.Rows;
      for (int i = 0; i < dimension; i++) {
        coordinates[i, 0] = 10 * Math.Cos(rad * i);
        coordinates[i, 1] = 10 * Math.Sin(rad * i);
      }
      double epsg = 0.0000000001;
      double epsf = 0;
      double epsx = 0;
      int maxits = 1000;
      alglib.mincgstate state = null;
      alglib.mincgreport rep;

      for (int iterations = 0; iterations < 200; iterations++) {
        for (int i = 0; i < dimension; i++) {
          double[] c = new double[] { coordinates[i, 0], coordinates[i, 1] };

          if (iterations == 0 && i == 0) {
            alglib.mincgcreate(c, out state);
            alglib.mincgsetcond(state, epsg, epsf, epsx, maxits);
          } else {
            alglib.mincgrestartfrom(state, c);
          }
          alglib.mincgoptimize(state, func, null, new Info(coordinates, distances, i));
          alglib.mincgresults(state, out c, out rep);

          coordinates[i, 0] = c[0];
          coordinates[i, 1] = c[1];
        }
      }
      error = 0;
      for (int i = 0; i < dimension - 1; i++) {
        for (int j = i + 1; j < dimension; j++) {
          if (distances[i, j] != 0) {
            double dx = coordinates[i, 0] - coordinates[j, 0];
            double dy = coordinates[i, 1] - coordinates[j, 1];
            double d = Math.Sqrt(dx * dx + dy * dy);
            error += ((distances[i, j] - d) * (distances[i, j] - d)) / (distances[i, j] * distances[i, j]);
          }
        }
      }
      return coordinates;
    }

    public static void func(double[] x, ref double func, double[] grad, object obj) {
      func = 0; grad[0] = 0; grad[1] = 0;
      Info info = (obj as Info);
      for (int i = 0; i < x.Length; i++) {
        if (i != info.Row) {
          double d = (x[0] - info.Coordinates[i, 0])
                   * (x[0] - info.Coordinates[i, 0])
                   + (x[1] - info.Coordinates[i, 1])
                   * (x[1] - info.Coordinates[i, 1]);
          func += (d - info.Distances[info.Row, i]) * (d - info.Distances[info.Row, i]);
          grad[0] += 2 * x[0] - 2 * info.Coordinates[i, 0];
          grad[1] += 2 * x[1] - 2 * info.Coordinates[i, 1];
        }
      }
    }

    private class Info {
      public DoubleMatrix Coordinates { get; set; }
      public DoubleMatrix Distances { get; set; }
      public int Row { get; set; }

      public Info(DoubleMatrix c, DoubleMatrix d, int r) {
        Coordinates = c;
        Distances = d;
        Row = r;
      }
    }
  }
}